#!/usr/bin/perl
# Create a report from a CVS commit
# Yes, this has some big inefficiencies in it.  I chose readability over
# raw power.
# $Fuller: myprograms/cvsmail/cvsmail.pl,v 1.1.1.2 2002/10/31 11:44:34 fullermd Exp $

use strict;
use Text::Tabs;

use lib $ENV{CVSROOT};
use CVSROOT::cvsmail;

my $CVSROOT = $ENV{'CVSROOT'};
die "No CVSROOT defined!" unless $CVSROOT;

umask(077);

# First try and parse out what files and revision's we're dealing with
my $infostr = $ARGV[0];	# Actually argv[1], stupid Perl...
chomp $infostr;

# Adding a dir?  Just silently drop out
if($infostr =~ /New\ directory$/)
{
	# Clean up and die quietly
	&cvsmail::clean_tmpfiles();
	exit(0);
}

# Are we an import?  This is going to take special handling...
if($infostr =~ /Imported\ sources$/)
{
	# Don't need to worry about multiple directories in this case.  Just
	# run with it.
	my $hdrline = &cvsmail::build_body_header();
	my @text;

	push @text, $hdrline;
	push @text, "";
	while(<STDIN>)
	{	chomp; push @text, "  $_";	}

	&cvsmail::send_mail(1, @text);
	&cvsmail::clean_tmpfiles();
	exit(0);
}

# If we get here, we've got a plain old ordinary run-of-the-mill commit.

# Set vars
my ($directory, @fileinfo) = split " ", $infostr;

my @path = split('/', $directory);
my $dir;
if($#path == 0)
{	$dir = ".";	}
else
{
	$dir = join('/', @path[1..$#path]);
}

$dir = $dir . "/";

# Tear up @fileinfo
my @filenames;
foreach (@fileinfo)
{
	my ($filename,$orev,$nrev) = split ",", $_;
	push @filenames, $filename;
	$cvsmail::orev{"$filename"} = $orev;
	$cvsmail::nrev{"$filename"} = $nrev;
}

# Build up our subject line
&cvsmail::append_line($cvsmail::FILE_SUBJECT,
		"$directory " . join(" ", sort @filenames));

# Biguglystatemachine for fiddling with the body data
# This owes heavily to the FreeBSD log_accum.pl
my $state = $cvsmail::STATE_NONE;
my @added_files;
my @changed_files;
my @removed_files;
my @log_lines;
my $tag;
while(<STDIN>)
{
	s/[ \t\n]+$//;		# Clear trailing space

	# Get revision tags if available
	if (/^Revision\/Branch:(.*)/)   { $tag = $1;     next; }
	if (/^[ \t]+Tag: (.*)/)     { $tag = $1;     next; }

	# Check for any state changes
	unless ($state == $cvsmail::STATE_LOG)
	{
		if (/^Modified Files/)  { $state = $cvsmail::STATE_CHANGED; next; }
		if (/^Added Files/) { $state = $cvsmail::STATE_ADDED; next; }
		if (/^Removed Files/)   { $state = $cvsmail::STATE_REMOVED; next; }
		if (/^Log Message/) { $state = $cvsmail::STATE_LOG; next; }
	}

	# If we haven't entered something real, flip back
	next if $state == $cvsmail::STATE_NONE;

	# If we're in the midst of the log message, suck it in
	if($state == $cvsmail::STATE_LOG)
	{
		push @log_lines, $_;
	}

	# Collect info about the files in the various states
	my @files = split;
	push @changed_files, @files if $state == $cvsmail::STATE_CHANGED;
	push @added_files, @files if $state == $cvsmail::STATE_ADDED;
	push @removed_files, @files if $state == $cvsmail::STATE_REMOVED;
}
# Done parsing STDIN stuff.

# Strip leading and trailing blank lines from the log message.
# Compress multiple blank lines in the body of the message down to a
# single blank line.
# Convert tabs to spaces, so that when we indent the email message and
# log file everything still lines up.
my $log_message = join "\n", @log_lines;
$log_message =~ s/\n{3,}/\n\n/g;
$log_message =~ s/^\n+//;
$log_message =~ s/\n+$//;
@log_lines = expand(split /\n/, $log_message);

# See if this log message has already been seen (this allows proper
# handling of multi-directory commits with same OR different log messages
# in each directory.  A bit obscure, and rather annoying, IMO, but it's
# the only rational way to handle it without getting *REALLY*
# complicated.
my $message_index;
for ($message_index = 0; ; $message_index++) {
	last unless -e "$cvsmail::FILE_LOGMSG.$message_index";

	my @text = &cvsmail::suck_file("$cvsmail::FILE_LOGMSG.$message_index");
	last unless @text;
	last if "@log_lines" eq "@text";
}


# Spit out whatever info we have
# Filenames and log messages
&cvsmail::file_filenames("$cvsmail::FILE_ADDED.$message_index",
		$dir, @added_files);
&cvsmail::file_filenames("$cvsmail::FILE_CHANGED.$message_index",
		$dir, @changed_files);
&cvsmail::file_filenames("$cvsmail::FILE_REMOVED.$message_index",
		$dir, @removed_files);
&cvsmail::blow_file("$cvsmail::FILE_LOGMSG.$message_index", @log_lines);

# If we have a branch, note it
if($tag)
{
	open FILE, "> $cvsmail::FILE_TAG";
	print FILE $tag;
	close FILE;
}

# Ferret out and file change amounts
&cvsmail::change_summary("$cvsmail::FILE_SUMMARY.$message_index",
		$directory, "ADDED", @added_files);
&cvsmail::change_summary("$cvsmail::FILE_SUMMARY.$message_index",
		$directory, "CHANGED", @changed_files);
&cvsmail::change_summary("$cvsmail::FILE_SUMMARY.$message_index",
		$directory, "REMOVED", @removed_files);

# Are we the last directory?
if (-e $cvsmail::FILE_LASTDIR)
{
	# Cheat so we don't have to write another function to get only a
	# single line.  Ho hum.  Lazy SOB, that's me!
	@_ = cvsmail::suck_file($cvsmail::FILE_LASTDIR);
	$_ = $_[0];
	
	# XXX This shouldn't be necessary, I think
	#my $tmploc = $directory;
	#$tmploc =! s,([^a-zA-Z0-9_/]),\\$1,g;
	# May sub $tmploc for $directory below?
	unless (grep(/$directory$/, $_))
	{
		print "More commits to come...\n";
		exit(0);
	}
}


# If we get here, this is the last dir, so it's time to assemble the
# mailing and blow it out.
my $i = 0;
while(-e "$cvsmail::FILE_LOGMSG.$i")
{
	my @log_msg = &cvsmail::build_body_header();

	# Note tag if we have one
	if (-e "$cvsmail::FILE_TAG")
	{
		open FILE, "$cvsmail::FILE_TAG";
		$tag = <FILE>;
		chomp $tag;
		close FILE;
		my $inline = sprintf("%-20s(Branch: $tag)","");
		push @log_msg, $inline;
	}
	else
	{	push @log_msg, "";	}

	# Put the filename summary types in
	my @mod_lines = &cvsmail::suck_file("$cvsmail::FILE_CHANGED.$i");
	push @log_msg, &cvsmail::format_lists("Modified", @mod_lines)
			if @mod_lines;

	my @add_lines = &cvsmail::suck_file("$cvsmail::FILE_ADDED.$i");
	push @log_msg, &cvsmail::format_lists("Added", @add_lines)
			if @add_lines;

	my @rem_lines = &cvsmail::suck_file("$cvsmail::FILE_REMOVED.$i");
	push @log_msg, &cvsmail::format_lists("Removed", @rem_lines)
			if @rem_lines;

	# Smack on the log message
	my @msg_lines = &cvsmail::suck_file("$cvsmail::FILE_LOGMSG.$i");
	push @log_msg, "  Log:", (map { "  $_" } @msg_lines)
			if @msg_lines;

	# Squit the delta summaries on
	if (-e "$cvsmail::FILE_SUMMARY.$i")
	{
		push @log_msg, "  ", map {"  $_"}
				cvsmail::format_summaries("$cvsmail::FILE_SUMMARY.$i");
	}

	# And send it out!
	&cvsmail::send_mail(0, @log_msg);

	# This will sent 1 mail per different log message.
	$i++;
}

&cvsmail::clean_tmpfiles();
exit(0);
# Fin!
