#!/DO/NOT/EXECUTE
# Underlying module for Matt's CVS mailing scripts
# Much material shamelessly cribbed from the FreeBSD Project's CVSROOT
# scripts.  Thanks, guys  :)
# $Fuller: myprograms/cvsmail/cvsmail.pm,v 1.6 2002/11/24 08:51:48 fullermd Exp $

package cvsmail;
use strict;
use vars qw(
	$MAILTO $PGID $TMPDIR $TMPFILEBASE $FILE_LASTDIR $FILE_SUBJECT
	$FILE_LOGMSG $FILE_ADDED $FILE_CHANGED $FILE_REMOVED $FILE_SUMMARY
	$FILE_TAG $PERP $IDHEADER $NO_UNEXPAND $UNEXPAND_ID $SENDMAIL
	$FROM_HOST $P_CVS $P_MV $P_HOSTNAME $DEBUG $STATE_NONE $STATE_CHANGED
	$STATE_ADDED $STATE_REMOVED $STATE_LOG $CUSTOM_HDR $CVSMAIL_VERSION
	%orev %nrev
);

my $CVSROOT = $ENV{'CVSROOT'} or die "Can't determine \$CVSROOT!";

################
# Variables
################

# Process group id; useful as a unique number
$PGID = getpgrp();

# Temporary file location
$TMPDIR = "/tmp/cvsmail";

# Temporary file name form
# Watch redeclaring this!
$TMPFILEBASE = "cvs$PGID";

# File names
# Probably don't want to mess with these.
$FILE_LASTDIR = "$TMPDIR/$TMPFILEBASE.lastdir";
$FILE_SUBJECT = "$TMPDIR/$TMPFILEBASE.subject";
$FILE_LOGMSG  = "$TMPDIR/$TMPFILEBASE.logmsg";
$FILE_ADDED   = "$TMPDIR/$TMPFILEBASE.added";
$FILE_CHANGED = "$TMPDIR/$TMPFILEBASE.changed";
$FILE_REMOVED = "$TMPDIR/$TMPFILEBASE.removed";
$FILE_SUMMARY = "$TMPDIR/$TMPFILEBASE.summary";
$FILE_TAG     = "$TMPDIR/$TMPFILEBASE.tag";

# Who's to blame
$PERP = $ENV{"LOGNAME"} || $ENV{'USER'} || getlogin ||
		(getpwuid($<))[0] || sprintf("uid#%d",$<);

# Extra variables
$STATE_NONE    = 0;
$STATE_CHANGED = 1;
$STATE_ADDED   = 2;
$STATE_REMOVED = 3;
$STATE_LOG     = 4;

# Our version
$CVSMAIL_VERSION = "1.4";


# Load config settings
eval { require "$CVSROOT/CVSROOT/cvsmail.cfg" };

# Sanity check
if($MAILTO eq "")
{
	print "**ERROR -> No \$MAILTO specified!\n";
	print "**ERROR -> Edit the cvsmail.cfg file like the\n";
	print "**ERROR -> instructions said, fool!\n";
	exit(0);
}

################
# Functions
################

# Clean up temporary files
sub clean_tmpfiles
{
	# We could do this nifty-like, but it's so much easier to just cheat
	# a bit and let the shell do it.
	system("rm -f $TMPDIR/$TMPFILEBASE.*");
}


# Get longest value in a list
sub longest_in_list
{
	my @values = @_;

	# Now this is a fucking absurd algorithm.  Oh well, it's Perl anyway,
	# what's one more totally nutbar Anti-Efficient mess?  At least it's
	# not going to be a very long list in 99.999% of commits.
	my @sorted = sort { $b <=> $a } map { length $_ } @values;
	return $sorted[0];
}


# Print the "header" line in the body of the message
sub build_body_header
{
	my $datestr = `/bin/date +"%Y/%m/%d %H:%M:%S %Z"`;
	chomp $datestr;
	my $header = sprintf("%-16s    %s", $PERP, $datestr);
	return $header;
}


# Add a line to a file
sub append_line
{
	my $filename = shift;
	my $line = shift;

	open FILE, ">>$filename";
	print FILE "$line\n";
	close FILE;
}


# Read a file into an array
sub suck_file
{
	my $filename = shift;
	my @text;

	if(open FILE, "<$filename")
	{
		while(<FILE>)
		{
			chomp;
			push @text, $_;
		}
		close FILE;
	}

	return @text;
}


# Write an array info a file
sub blow_file
{
	my $filename = shift;
	my @lines = @_;

	open FILE, ">$filename";
	print FILE map { "$_\n" } @lines;
	close FILE;
}


# Put a list of files of a specific type and log message info a file.
# This is part of aggregating all the crap.
sub file_filenames
{
	my $filename = shift;
	my $dir = shift;
	my @files = @_;

	return unless @files;

	open FILE, ">>$filename";

	print FILE "$dir\n";
	print FILE map { "$_\n" } @files;
	close FILE;
}


# Count total lines in a given revision
sub count_lines_in_rev
{
	my $file = shift;
	my $rev = shift;

	my $lines = 0;
	# I'm not entirely sure this SHOULD work, but it does
	open(RCS, "-|") ||
			exec $P_CVS, '-Qn', 'update', '-p', "-r$rev", $file;
	while(<RCS>)
	{	$lines++;	}
	close(RCS);

	return $lines;
}


# Summarize changes in various types of files
# New files
sub change_summary
{
	my $outfile = shift;
	my $directory = shift;
	my $type = shift;
	my @filenames = @_;


	my $stat;
	my $rev;
	my $lines;
	my $delta;

	foreach my $file (@filenames)
	{
		next unless $file;
		if($type eq "ADDED")
		{
			$stat = "new";
			$rev = $nrev{"$file"};
			$lines = &count_lines_in_rev($file, $rev);
			$delta = "+$lines -0";
		}
		elsif($type eq "REMOVED")
		{
			$stat = "dead";
			$rev = $orev{"$file"};
			$lines = &count_lines_in_rev($file, $rev);
			$delta = "+0 -$lines";
		}
		else	#if($type eq "CHANGED")
		{
			$stat = "";
			$rev = $nrev{"$file"};
			open(RCS, "-|") ||
					exec $P_CVS, '-Qn', 'log', "-r$rev", $file;
			while(<RCS>)
			{
				if(/^date:.*lines:\s(.*)$/)
				{
					$delta = $1;
					last;
				}
			}
			close(RCS);
		}

		&append_line($outfile, "$rev,$delta,$stat,$directory/$file");
	}
}


# Format filenames within 1 directory
sub format_names
{
	my $dir = shift;
	my @files = @_;

	my $indent = length($dir);
	if($indent < 20)
	{	$indent = 20;	}

	my $format = "    %-" . sprintf("%d", $indent) . "s ";
	my @lines = (sprintf($format, $dir));

	foreach my $file (@files)
	{
		# 66 == arbitrary line length limitation
		if(length($lines[$#lines]) + length($file) > 66)
		{
			$lines[++$#lines] = sprintf("$format", "", "");
		}
		$lines[$#lines] .= $file . " ";
	}

	return @lines;
}


# Format lists of files in the mail
sub format_lists
{
	my $header = shift;
	my @lines = @_;

	my $lastdir = '';

	push my @text, "  $header files:";

	my @files;
	foreach my $line (@lines)
	{
		if($line =~ /.*\/$/)
		{
			# We've hit the beginning of a directory.  If $lastdir is
			# defined, we already have a directory being gathered, so
			# format and spit it out.  Else, this is the first dir, so
			# designate it and get on with the gathering.
			push @text, &format_names($lastdir, sort @files) if $lastdir;
			@files = ();

			$lastdir = $line;
			$lastdir =~ s,/$,,;
		}
		else
		{
			push @files, $line;
		}
	}

	push @text, &format_names($lastdir, sort @files);

	return @text;
}


# Format "summary" information
sub format_summaries
{
	my $file = shift;

	my @revs;
	my @deltas;
	my @statae;
	my @files;

	# Parse the file
	open FILE, $file;
	while(<FILE>)
	{
		chomp;
		my ($r, $d, $s, $f) = split(/,/, $_, 4);
		push @revs, $r;
		push @deltas, $d;
		push @statae, $s;
		push @files, $f;
	}
	close FILE;

	# Format it all
	my $r_max = longest_in_list("Revision", @revs) + 2;
	my $d_max = longest_in_list("Changes", @deltas) + 2;

	my $fmt = "%-" . $r_max . "s%-" . $d_max . "s%s";
	push my @text, sprintf $fmt, "Revision", "Changes", "Path";
	my @order = sort { $files[$a] cmp $files[$b] } (0 .. $#revs);
	foreach (@order)
	{
		my $file = $files[$_];
		my $status = $statae[$_];
		$file .= " ($status)" if $status;
		push @text, sprintf $fmt, $revs[$_], $deltas[$_], $file;
	}

	return @text;
}


# Send the whole mail
sub send_mail
{
	my $isimport = shift;
	my @text = @_;

	if($FROM_HOST eq "")
	{
		$FROM_HOST = `$P_HOSTNAME`;
		chomp $FROM_HOST;
	}

	my $subject = "Subject: cvs commit:";
	my @subj = &suck_file($FILE_SUBJECT);	

	# Make the Subject: continutation rational and split into
	# evenly-sized lines.  Snarfed from FreeBSD scripts verbatim.
	my $subjlines = 0;
	my $subjwords = 0;
	my @email;
	LINE: foreach my $line (@subj)
	{
		foreach my $word (split(/ /, $line))
		{
			if ($subjwords > 2 && length("$subject $word") > 75)
			{
				if ($subjlines > 2)
				{	$subject .= " ...";	}
				push @email, $subject;
				if ($subjlines > 2)
				{
					$subject = "";
					last LINE;
				}
				# rfc822 continuation line
				$subject = "        ";
				$subjwords = 0;
				$subjlines++;
			}
			$subject .= " " . $word;
			$subjwords++;
		}
	}
	if($isimport==1)
	{ $subject = "Subject: $ARGV[0]"; }
	chomp $subject;
	push @email, $subject if $subject;

	# Assemble the messages
	my @snd_msg;
	push @snd_msg, "From: $PERP\@$FROM_HOST";
	push @snd_msg, "To: $MAILTO";
	push @snd_msg, @email if @email;
	push @snd_msg, "X-Mailer: CVSmail $CVSMAIL_VERSION";
	push @snd_msg, $CUSTOM_HDR if($CUSTOM_HDR ne "");
	push @snd_msg, "";
	push @snd_msg, @text;

	# Debug
	if($DEBUG==1)
	{
		print "Message being sent:\n***\n";
		print map { "$_\n" } @snd_msg;
		print "***\n";
	}

	print "Sending message to $MAILTO...  ";
	open MAIL, "| $SENDMAIL $MAILTO" or die "Unable to pipe to $SENDMAIL";
	print MAIL map { "$_\n" } @snd_msg;
	close MAIL;
	print "Done.\n";
}

# We're a module, return true
1;
# Fin
